
/*
 * one line to give the Gauche-archive and an idea of what it does.
 * Copyright (C) 2018 aaauuue
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "archive_proc.h"

int find_slot_index( char *name, ScmObj obj )
{
  ScmObj p;
  ScmObj src;
  ScmObj sym = Scm_MakeSymbol( SCM_STRING( SCM_MAKE_STR( name ) ), 1 );
  int i      = 0;

  SCM_FOR_EACH( p, Scm_ClassOf( obj )->directSlots )
  {
    src = SCM_CAR( SCM_CAR( p ) );
    /* Scm_Printf( port, "[%S,%S]\n", src, sym ); */
    if( SCM_EQ( src, sym ) ) {
      return i;
    }
    i++;
  }

  return -1;
}

ScmObj find_name( struct archive_entry *entry, ScmObj lst )
{
  ScmObj p;
  const char *src  = NULL;
  const char *dest = NULL;
  int index        = -1;

  src = archive_entry_pathname_utf8( entry );
  /* fprintf( stderr, "src:[%s]\n", src ); */
  if( src == NULL ) {
    return SCM_FALSE;
  }
  SCM_FOR_EACH( p, lst )
  {
    index = find_slot_index( "name", SCM_CAR( p ) );
    dest  = SCM_CSTR( Scm_InstanceSlotRef( SCM_CAR( p ), index ) );
    /* fprintf( stderr, "src:[%s] dest:[%s]\n", src, dest ); */
    if( strcmp( src, dest ) == 0 ) {
      return p;
    }
  }

  return SCM_FALSE;
}

ScmObj open_read_archive_from_filename( ScmObj obj, ScmObj name )
{
  Scm_libarchive_archive_base *a = NULL;
  int r                          = ARCHIVE_OK;
  int index                      = -1;

  index = find_slot_index( "archive-core", obj );
  a     = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( obj, index ) );

  a->rw_type  = GAUCHE_ARCHIVE_READ;
  a->src_type = ARCHIVE_SRC_FILE;
  a->obj      = archive_read_new();
  archive_read_support_filter_all( a->obj );
  archive_read_support_format_all( a->obj );
  a->name = strdup( SCM_CSTR( name ) );
  r       = archive_read_open_filename( a->obj, a->name, BUFSIZ );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( a->obj ), SCM_NIL );
  }

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj open_read_archive_from_memory( ScmObj obj, ScmObj port )
{
  Scm_libarchive_archive_base *a = NULL;
  ScmPort *p                     = SCM_PORT( port );
  int r                          = ARCHIVE_OK;
  int index                      = -1;

  if( SCM_PORT_DIR( p ) != SCM_PORT_OUTPUT ||
      SCM_PORT_TYPE( p ) != SCM_PORT_OSTR ) {
    return AR_ERROR( -1, "<output-string-port> required.", SCM_NIL );
  }

  index = find_slot_index( "archive-core", obj );
  a     = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( obj, index ) );

  a->rw_type  = GAUCHE_ARCHIVE_READ;
  a->src_type = ARCHIVE_SRC_MEMORY;
  a->buflen   = ( size_t )Scm_DStringSize( &( p->src.ostr ) );
  a->buf      = ( char * )Scm_DStringGetz( &( p->src.ostr ) );
  a->obj      = archive_read_new();
  archive_read_support_filter_all( a->obj );
  archive_read_support_format_all( a->obj );
  r = archive_read_open_memory( a->obj, a->buf, a->buflen );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( a->obj ), SCM_NIL );
  }

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj close_read_archive( ScmObj obj )
{
  Scm_libarchive_archive_base *a = NULL;
  int index                      = -1;

  index = find_slot_index( "archive-core", obj );
  a     = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( obj, index ) );

  if( a->obj ) {
    switch( a->rw_type ) {
    case GAUCHE_ARCHIVE_READ:
      archive_read_close( a->obj );
      archive_read_free( a->obj );
      a->obj = NULL;
      break;

    case GAUCHE_ARCHIVE_WRITE:
      archive_write_close( a->obj );
      archive_write_free( a->obj );
      a->obj = NULL;

    default:
      break;
    }
  }

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj free_read_archive( ScmObj obj )
{
  libarchive_archive_base_finalize( obj, NULL );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj is_read_archive_open( ScmObj obj )
{
  Scm_libarchive_archive_base *a = NULL;
  int index                      = -1;

  index = find_slot_index( "archive-core", obj );
  a     = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( obj, index ) );

  if( a->obj != NULL ) {
    return AR_ERROR( 1, "archive opened.", SCM_NIL );
  }

  return AR_ERROR( 0, "archive closed.", SCM_TRUE );
}

ScmObj get_entry( ScmObj archive, ScmObj entry )
{
  Scm_libarchive_archive_base *a = NULL;
  struct archive_entry *_entry   = NULL;
  int r                          = ARCHIVE_OK;
  int index                      = -1;

#ifdef HAVE_SYS_STAT_H
  const struct stat *st = NULL;
#endif /* #ifdef HAVE_SYS_STAT_H */

  index = find_slot_index( "archive-core", archive );
  a     = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( archive, index ) );

  if( !a->obj ) {
    return AR_ERROR( -1, "Archive isn't open yet.", SCM_NIL );
  }

  r = archive_read_next_header( a->obj, &_entry );
  if( r == ARCHIVE_EOF ) {
    return Scm_Values4( SCM_FALSE, SCM_FALSE, SCM_FALSE, SCM_FALSE );
  }
  if( r == ARCHIVE_OK ) {
    index = find_slot_index( "name", entry );
    Scm_InstanceSlotSet(
        entry, index,
        SCM_MAKE_STR_COPYING( archive_entry_pathname_utf8( _entry ) ) );

#ifdef HAVE_SYS_STAT_H
    st = archive_entry_stat( _entry );

    index = find_slot_index( "isdir", entry );
    if( S_ISREG( st->st_mode ) ) {
      Scm_InstanceSlotSet( entry, index, SCM_FALSE );
    } else if( S_ISDIR( st->st_mode ) ) {
      Scm_InstanceSlotSet( entry, index, SCM_TRUE );
    }
    index = find_slot_index( "uid", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_uid ) );
    index = find_slot_index( "gid", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_gid ) );
    index = find_slot_index( "mode", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_mode ) );
    index = find_slot_index( "atime", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_atim.tv_sec ) );
    index = find_slot_index( "ctime", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_ctim.tv_sec ) );
    index = find_slot_index( "mtime", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_mtim.tv_sec ) );
    index = find_slot_index( "size", entry );
    Scm_InstanceSlotSet( entry, index, SCM_MAKE_INT( st->st_size ) );
#endif /* #ifdef HAVE_SYS_STAT_H */

    return AR_ERROR( 0, "", SCM_TRUE );
  }

  return AR_ERROR( r, archive_error_string( a->obj ), SCM_NIL );
}

ScmObj read_data( ScmObj archive, ScmObj entry )
{
  int flags                      = 0;
  int r                          = 0;
  int index                      = -1;
  long len                       = 0;
  char buf[BUFSIZ]               = { '\0' };
  Scm_libarchive_archive_base *a = NULL;
  ScmPort *sport                 = NULL;

  index = find_slot_index( "archive-core", archive );
  a     = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( archive, index ) );

  index = find_slot_index( "data", entry );
  sport = SCM_PORT( Scm_InstanceSlotRef( entry, index ) );
  len   = archive_read_data( a->obj, buf, sizeof( buf ) );
  while( len > 0 ) {
    Scm_Putz( buf, len, sport );
    len = archive_read_data( a->obj, buf, sizeof( buf ) );
  }
  if( len < 0 ) {
    return AR_ERROR( -1, archive_error_string( a->obj ), SCM_NIL );
  }

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj extract_file_to_disk( ScmObj archive, ScmObj entry_list )
{
  int flags                           = 0;
  int r                               = 0;
  long len                            = 0;
  char buf[BUFSIZ]                    = { '\0' };
  struct archive *ext                 = NULL;
  struct archive *a                   = NULL;
  struct archive_entry *entry_iter    = NULL;
  Scm_libarchive_archive_base *a_base = NULL;
  int index                           = -1;

  index  = find_slot_index( "archive-core", archive );
  a_base = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( archive, index ) );

  flags = ARCHIVE_EXTRACT_TIME;
  flags |= ARCHIVE_EXTRACT_PERM;
  flags |= ARCHIVE_EXTRACT_ACL;
  flags |= ARCHIVE_EXTRACT_FFLAGS;

  a = archive_read_new();
  archive_read_support_filter_all( a );
  archive_read_support_format_all( a );

  switch( a_base->src_type ) {
  case ARCHIVE_SRC_FILE:
    if( ( r = archive_read_open_filename( a, a_base->name, BUFSIZ ) ) ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    break;
  case ARCHIVE_SRC_MEMORY:
    if( ( r = archive_read_open_memory( a, a_base->buf, a_base->buflen ) ) ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    break;
  default:
    break;
  }

  ext = archive_write_disk_new();
  archive_write_disk_set_options( ext, flags );
  archive_write_disk_set_standard_lookup( ext );

  for( ;; ) {
    r = archive_read_next_header( a, &entry_iter );
    if( r == ARCHIVE_EOF ) {
      break;
    }
    if( r < ARCHIVE_OK ) {
      archive_read_close( a );
      archive_read_free( a );
      archive_write_close( ext );
      archive_write_free( ext );
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    if( r < ARCHIVE_WARN ) {
      archive_read_close( a );
      archive_read_free( a );
      archive_write_close( ext );
      archive_write_free( ext );
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }

    if( find_name( entry_iter, entry_list ) != SCM_FALSE ) {
      r = archive_write_header( ext, entry_iter );
      if( r < ARCHIVE_OK ) {
        archive_read_close( a );
        archive_read_free( a );
        archive_write_close( ext );
        archive_write_free( ext );
        return AR_ERROR( r, archive_error_string( ext ), SCM_NIL );
      }
      if( archive_entry_size( entry_iter ) > 0 ) {
        len = archive_read_data( a, buf, sizeof( buf ) );
        while( len > 0 ) {
          if( archive_write_data( ext, buf, len ) != len ) {
            archive_read_close( a );
            archive_read_free( a );
            archive_write_close( ext );
            archive_write_free( ext );
            return AR_ERROR( -1, archive_error_string( ext ), SCM_NIL );
          }
          len = archive_read_data( a, buf, sizeof( buf ) );
        }
        if( len < 0 ) {
          archive_read_close( a );
          archive_read_free( a );
          archive_write_close( ext );
          archive_write_free( ext );
          return AR_ERROR( -1, archive_error_string( a ), SCM_NIL );
        }
      }

      r = archive_write_finish_entry( ext );
      if( r < ARCHIVE_OK ) {
        archive_read_close( a );
        archive_read_free( a );
        archive_write_close( ext );
        archive_write_free( ext );
        return AR_ERROR( -1, archive_error_string( ext ), SCM_NIL );
      }
      if( r < ARCHIVE_WARN ) {
        archive_read_close( a );
        archive_read_free( a );
        archive_write_close( ext );
        archive_write_free( ext );
        return AR_ERROR( -1, archive_error_string( ext ), SCM_NIL );
      }
    }
  }

  archive_read_close( a );
  archive_read_free( a );
  archive_write_close( ext );
  archive_write_free( ext );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj extract_file_to_memory( ScmObj archive, ScmObj entry_list )
{
  int flags                           = 0;
  int r                               = 0;
  int index                           = -1;
  long len                            = 0;
  char buf[BUFSIZ]                    = { '\0' };
  struct archive *a                   = NULL;
  struct archive_entry *entry_iter    = NULL;
  Scm_libarchive_archive_base *a_base = NULL;
  ScmPort *oport                      = NULL;
  ScmObj entry;

  index  = find_slot_index( "archive-core", archive );
  a_base = SCM_LIBARCHIVE_ARCHIVE_BASE( Scm_InstanceSlotRef( archive, index ) );

  flags = ARCHIVE_EXTRACT_TIME;
  flags |= ARCHIVE_EXTRACT_PERM;
  flags |= ARCHIVE_EXTRACT_ACL;
  flags |= ARCHIVE_EXTRACT_FFLAGS;

  a = archive_read_new();
  archive_read_support_filter_all( a );
  archive_read_support_format_all( a );
  switch( a_base->src_type ) {
  case ARCHIVE_SRC_FILE:
    if( ( r = archive_read_open_filename( a, a_base->name, BUFSIZ ) ) ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    break;
  case ARCHIVE_SRC_MEMORY:
    if( ( r = archive_read_open_memory( a, a_base->buf, a_base->buflen ) ) ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    break;
  default:
    break;
  }

  for( ;; ) {
    r = archive_read_next_header( a, &entry_iter );
    if( r == ARCHIVE_EOF ) {
      break;
    }
    if( r < ARCHIVE_OK ) {
      archive_read_close( a );
      archive_read_free( a );
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    if( r < ARCHIVE_WARN ) {
      archive_read_close( a );
      archive_read_free( a );
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }

    entry = find_name( entry_iter, entry_list );
    if( entry != SCM_FALSE ) {
      if( archive_entry_size( entry_iter ) > 0 ) {
        index = find_slot_index( "data", SCM_CAR( entry ) );
        oport = SCM_PORT( Scm_InstanceSlotRef( SCM_CAR( entry ), index ) );
        len   = archive_read_data( a, buf, sizeof( buf ) );
        while( len > 0 ) {
          Scm_Putz( buf, len, SCM_PORT( oport ) );
          len = archive_read_data( a, buf, sizeof( buf ) );
        }
        if( len < 0 ) {
          return AR_ERROR( -1, archive_error_string( a ), SCM_NIL );
        }
      }
    }
  }

  archive_read_close( a );
  archive_read_free( a );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj extract_all_file( ScmObj archive )
{
  int flags                        = 0;
  int r                            = 0;
  long len                         = 0;
  char buf[BUFSIZ]                 = { '\0' };
  Scm_libarchive_archive_base *p   = NULL;
  struct archive *ext              = NULL;
  struct archive *a                = NULL;
  struct archive_entry *entry_iter = NULL;

  p = SCM_LIBARCHIVE_ARCHIVE_BASE( archive );
  if( p->rw_type != GAUCHE_ARCHIVE_READ ) {
    return AR_ERROR( -1, "reader object is required.", SCM_NIL );
  }

  flags = ARCHIVE_EXTRACT_TIME;
  flags |= ARCHIVE_EXTRACT_PERM;
  flags |= ARCHIVE_EXTRACT_ACL;
  flags |= ARCHIVE_EXTRACT_FFLAGS;

  a = archive_read_new();
  archive_read_support_filter_all( a );
  archive_read_support_format_all( a );
  ext = archive_write_disk_new();
  archive_write_disk_set_options( ext, flags );
  archive_write_disk_set_standard_lookup( ext );

  switch( p->src_type ) {
  case ARCHIVE_SRC_FILE:
    if( ( r = archive_read_open_filename( a, p->name, BUFSIZ ) ) ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    break;
  case ARCHIVE_SRC_MEMORY:
    if( ( r = archive_read_open_memory( a, p->buf, p->buflen ) ) ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    break;
  default:
    break;
  }

  for( ;; ) {
    r = archive_read_next_header( a, &entry_iter );
    if( r == ARCHIVE_EOF ) {
      break;
    }
    if( r < ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    if( r < ARCHIVE_WARN ) {
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }

    r = archive_write_header( ext, entry_iter );
    if( r < ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( ext ), SCM_NIL );
    }
    if( archive_entry_size( entry_iter ) > 0 ) {
      len = archive_read_data( a, buf, sizeof( buf ) );
      while( len > 0 ) {
        if( archive_write_data( ext, buf, len ) != len ) {
          return AR_ERROR( -1, archive_error_string( ext ), SCM_NIL );
        }
        len = archive_read_data( a, buf, sizeof( buf ) );
      }
      if( len < 0 ) {
        return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
      }
    }

    r = archive_write_finish_entry( ext );
    if( r < ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( ext ), SCM_NIL );
    }
    if( r < ARCHIVE_WARN ) {
      return AR_ERROR( r, archive_error_string( ext ), SCM_NIL );
    }
  }

  archive_read_close( a );
  archive_read_free( a );
  archive_write_close( ext );
  archive_write_free( ext );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj create_archive( ScmObj archive_name, ScmObj file_list,
                       ScmObj compress_format, ScmObj filter_format )
{
  struct archive *a           = NULL;
  struct archive_entry *entry = NULL;
  char buf[BUFSIZ]            = { '\0' };
  int len                     = 0;
  int fd                      = -1;
  int r                       = ARCHIVE_OK;
  const char *s               = NULL;
  ScmObj p;
#ifdef HAVE_SYS_STAT_H
  struct stat st;
#endif /* #ifdef HAVE_SYS_STAT_H */

  a = archive_write_new();
  archive_write_set_format( a, Scm_GetInteger( compress_format ) );
  archive_write_add_filter( a, Scm_GetInteger( filter_format ) );
  r = archive_write_open_filename( a, SCM_CSTR( archive_name ) );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
  }

  entry = archive_entry_new();
  SCM_FOR_EACH( p, file_list )
  {
    archive_entry_clear( entry );
    s = SCM_CSTR( SCM_CAR( p ) );
    archive_entry_set_pathname( entry, s );
#ifdef HAVE_SYS_STAT_H
    stat( s, &st );
    archive_entry_set_size( entry, st.st_size );
    archive_entry_copy_stat( entry, &st );
    if( ( st.st_atim.tv_sec == 0 || st.st_ctim.tv_sec == 0 ) &&
        st.st_mtim.tv_sec != 0 ) {
      archive_entry_set_mtime( entry, st.st_mtim.tv_sec, 0 );
      archive_entry_set_atime( entry, st.st_mtim.tv_sec, 0 );
      archive_entry_set_ctime( entry, st.st_mtim.tv_sec, 0 );
    }
#endif /* #ifdef HAVE_SYS_STAT_H */
    r = archive_write_header( a, entry );
    if( r != ARCHIVE_OK ) {
      archive_entry_free( entry );
      archive_write_close( a );
      archive_write_free( a );
      return AR_ERROR( r, archive_error_string( a ), SCM_NIL );
    }
    fd  = open( s, O_RDONLY );
    len = read( fd, buf, BUFSIZ );
    while( len > 0 ) {
      archive_write_data( a, buf, len );
      len = read( fd, buf, BUFSIZ );
    }
    close( fd );
  }
  archive_entry_free( entry );

  archive_write_close( a );
  archive_write_free( a );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj copy_write_archive( ScmObj tmp_name, ScmObj archive_name,
                           ScmObj remove_file_list, ScmObj append_file_list,
                           ScmObj compress_format, ScmObj filter_format )
{
  struct archive *ar          = NULL;
  struct archive *aw          = NULL;
  struct archive_entry *entry = NULL;
  char buf[BUFSIZ]            = { '\0' };
  int len                     = 0;
  int fd                      = -1;
  int isdir                   = 0;
  int index                   = -1;
  int r                       = ARCHIVE_OK;
  const char *s               = NULL;
  ScmPort *oport              = NULL;
  ScmObj p;
  ScmObj scm_entry;
  ScmObj sym;
#ifdef HAVE_SYS_STAT_H
  struct stat st;
#endif /* #ifdef HAVE_SYS_STAT_H */

  aw = archive_write_new();
  archive_write_set_format( aw, Scm_GetInteger( compress_format ) );
  archive_write_add_filter( aw, Scm_GetInteger( filter_format ) );
  r = archive_write_open_filename( aw,
                                   Scm_GetString( SCM_STRING( tmp_name ) ) );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
  }

  ar = archive_read_new();
  archive_read_support_filter_all( ar );
  archive_read_support_format_all( ar );
  r = archive_read_open_filename(
      ar, Scm_GetString( SCM_STRING( archive_name ) ), BUFSIZ );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( ar ), SCM_NIL );
  }

  for( ;; ) {
    r = archive_read_next_header( ar, &entry );
    if( r == ARCHIVE_EOF ) {
      break;
    }

    scm_entry = find_name( entry, remove_file_list );
    if( scm_entry != SCM_FALSE ) { // found remove entry
      continue;
    } else { // copy from src to dest.
      r = archive_write_header( aw, entry );
      if( r != ARCHIVE_OK ) {
        return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
      }
      if( archive_entry_size( entry ) > 0 ) {
        len = archive_read_data( ar, buf, BUFSIZ );
        while( len > 0 ) {
          if( archive_write_data( aw, buf, len ) != len ) {
            return AR_ERROR( -1, "Size mismatch", SCM_NIL );
          }
          len = archive_read_data( ar, buf, BUFSIZ );
        }
        if( len < 0 ) {
          return AR_ERROR( -1, "Size mismatch", SCM_NIL );
        }
      }
    }
  }

  SCM_FOR_EACH( p, append_file_list )
  { // append new
    entry     = archive_entry_new();
    scm_entry = SCM_CAR( p );

    index = find_slot_index( "isdir", scm_entry );
    if( Scm_InstanceSlotRef( scm_entry, index ) == SCM_FALSE ) {
      archive_entry_set_mode( entry, 0744 | S_IFREG );
      isdir = 0;
    } else {
      archive_entry_set_mode( entry, 0744 | S_IFDIR );
      isdir = 1;
    }
    index = find_slot_index( "uid", scm_entry );
    archive_entry_set_uid(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    index = find_slot_index( "gid", scm_entry );
    archive_entry_set_gid(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    index = find_slot_index( "atime", scm_entry );
    archive_entry_set_atime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "ctime", scm_entry );
    archive_entry_set_ctime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "mtime", scm_entry );
    archive_entry_set_mtime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "size", scm_entry );
    archive_entry_set_size(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );
    index = find_slot_index( "name", scm_entry );
    archive_entry_copy_pathname(
        entry, Scm_GetString(
                   SCM_STRING( Scm_InstanceSlotRef( scm_entry, index ) ) ) );
    /*
     * fprintf( stderr, "[%s]\n", Scm_GetString( SCM_STRING( Scm_InstanceSlotRef( scm_entry, index ) ) ) );
     * fprintf( stderr, "[%s]\n", archive_entry_pathname_utf8( entry ) );
     */

    // write header
    r = archive_write_header( aw, entry );
    if( r != ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
    }

    if( isdir == 0 ) { /* file */
      index = find_slot_index( "type", scm_entry );
      sym   = Scm_InstanceSlotRef( scm_entry, index );
      index = find_slot_index( "data", scm_entry );
      oport = SCM_PORT( Scm_InstanceSlotRef( scm_entry, index ) );
      len   = SCM_DSTRING_SIZE( &( oport->src.ostr ) );

      if( COMPARE_SYMBOL( sym, "file" ) && len == 0 ) { /* header only */
        fd  = open( s, O_RDONLY );
        len = read( fd, buf, BUFSIZ );
        while( len > 0 ) {
          if( archive_write_data( aw, buf, len ) != len ) {
            AR_ERROR( -1, "Size mismatch", SCM_NIL );
          }
          len = read( fd, buf, BUFSIZ );
        }
        close( fd );
      } else {
        if( archive_write_data( aw, Scm_DStringGetz( &( oport->src.ostr ) ),
                                len ) != len ) {
          return AR_ERROR( -1, "Size mismatch", SCM_NIL );
        }
      }
    }

    archive_entry_free( entry );
  }

  archive_read_close( ar );
  archive_read_free( ar );
  archive_write_close( aw );
  archive_write_free( aw );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj write_archive_to_disk( ScmObj name, ScmObj file_list,
                              ScmObj compress_format, ScmObj filter_format,
                              ScmObj option_list )
{
  struct archive *aw          = NULL;
  struct archive_entry *entry = NULL;
  char buf[BUFSIZ]            = { '\0' };
  int len                     = 0;
  int index                   = -1;
  int name_index              = -1;
  int fd                      = -1;
  int isdir                   = 0;
  int r                       = ARCHIVE_OK;
  ScmObj p;
  ScmObj scm_entry;
  ScmObj sym;
  ScmPort *oport;

  aw = archive_write_new();
  archive_write_set_format( aw, Scm_GetInteger( compress_format ) );
  archive_write_add_filter( aw, Scm_GetInteger( filter_format ) );

  SCM_FOR_EACH( p, option_list )
  {
    /* fprintf( stderr, "option=[%s]\n", SCM_CSTR( SCM_CAR( p ) ) ); */
    r = archive_write_set_options( aw, SCM_CSTR( SCM_CAR( p ) ) );
    if( r != ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
    }
  }

  r = archive_write_open_filename( aw, SCM_CSTR( name ) );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
  }

  entry = archive_entry_new();
  SCM_FOR_EACH( p, file_list )
  {
    scm_entry = SCM_CAR( p );

    index      = find_slot_index( "name", scm_entry );
    name_index = index;
    archive_entry_copy_pathname(
        entry, Scm_GetString(
                   SCM_STRING( Scm_InstanceSlotRef( scm_entry, index ) ) ) );
    /* fprintf( stderr, "[%s]\n", Scm_GetString( SCM_STRING( Scm_InstanceSlotRef( scm_entry, index ) ) ) ); */
    /* fprintf( stderr, "[%s]\n", archive_entry_pathname_utf8( entry ) ); */

    index = find_slot_index( "isdir", scm_entry );
    if( Scm_InstanceSlotRef( scm_entry, index ) == SCM_FALSE ) {
      archive_entry_set_mode( entry, 0744 | S_IFREG );
      isdir = 0;
    } else {
      archive_entry_set_mode( entry, 0744 | S_IFDIR );
      isdir = 1;
    }
    index = find_slot_index( "uid", scm_entry );
    archive_entry_set_uid(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    index = find_slot_index( "gid", scm_entry );
    archive_entry_set_gid(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    index = find_slot_index( "atime", scm_entry );
    archive_entry_set_atime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "ctime", scm_entry );
    archive_entry_set_ctime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "mtime", scm_entry );
    archive_entry_set_mtime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "size", scm_entry );
    archive_entry_set_size(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    /* write header */
    r = archive_write_header( aw, entry );
    if( r != ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
    }
    if( isdir == 0 ) { /* file */
      index = find_slot_index( "type", scm_entry );
      sym   = Scm_InstanceSlotRef( scm_entry, index );
      index = find_slot_index( "data", scm_entry );
      oport = SCM_PORT( Scm_InstanceSlotRef( scm_entry, index ) );
      len   = SCM_DSTRING_SIZE( &( oport->src.ostr ) );

      if( COMPARE_SYMBOL( sym, "file" ) && len == 0 ) { /* header only */
        fd  = open( SCM_CSTR( Scm_InstanceSlotRef( scm_entry, name_index ) ),
                   O_RDONLY );
        len = read( fd, buf, BUFSIZ );
        while( len > 0 ) {
          if( archive_write_data( aw, buf, len ) != len ) {
            AR_ERROR( -1, "Size mismatch", SCM_NIL );
          }
          len = read( fd, buf, BUFSIZ );
        }
        close( fd );
      } else {
        if( archive_write_data( aw, Scm_DStringGetz( &( oport->src.ostr ) ),
                                len ) != len ) {
          return AR_ERROR( -1, "Size mismatch", SCM_NIL );
        }
      }
    }

    archive_entry_clear( entry );
  }
  archive_entry_free( entry );

  archive_write_close( aw );
  archive_write_free( aw );

  return AR_ERROR( 0, "", SCM_TRUE );
}

ScmObj write_archive_to_memory( ScmObj size, ScmObj file_list,
                                ScmObj compress_format, ScmObj filter_format,
                                ScmObj option_list )
{
  struct archive *aw          = NULL;
  struct archive_entry *entry = NULL;
  char buf[BUFSIZ]            = { '\0' };
  char *archive_buffer        = NULL;
  size_t archive_buffer_size  = 0;
  size_t used                 = 0;
  int len                     = 0;
  int index                   = -1;
  int isdir                   = 0;
  int name_index              = -1;
  int fd                      = -1;
  int r                       = ARCHIVE_OK;
  ScmObj p;
  ScmObj scm_entry;
  ScmObj sym;
  ScmObj u8;
  ScmPort *oport = NULL;

  archive_buffer_size = ( Scm_GetInteger( size ) * 2 );
  archive_buffer = ( char * )malloc( sizeof( char ) * archive_buffer_size );

  aw = archive_write_new();
  archive_write_set_format( aw, Scm_GetInteger( compress_format ) );
  archive_write_add_filter( aw, Scm_GetInteger( filter_format ) );

  SCM_FOR_EACH( p, option_list )
  {
    /* fprintf( stderr, "option=[%s]\n", SCM_CSTR( SCM_CAR( p ) ) ); */
    r = archive_write_set_options( aw, SCM_CSTR( SCM_CAR( p ) ) );
    if( r != ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
    }
  }

  r = archive_write_open_memory( aw, archive_buffer, archive_buffer_size,
                                 &used );
  if( r != ARCHIVE_OK ) {
    return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
  }

  entry = archive_entry_new();
  SCM_FOR_EACH( p, file_list )
  {
    scm_entry = SCM_CAR( p );

    index = find_slot_index( "isdir", scm_entry );
    if( Scm_InstanceSlotRef( scm_entry, index ) == SCM_FALSE ) {
      archive_entry_set_mode( entry, 0744 | S_IFREG );
      isdir = 0;
    } else {
      archive_entry_set_mode( entry, 0744 | S_IFDIR );
      isdir = 1;
    }
    index = find_slot_index( "uid", scm_entry );
    archive_entry_set_uid(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    index = find_slot_index( "gid", scm_entry );
    archive_entry_set_gid(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );

    index = find_slot_index( "atime", scm_entry );
    archive_entry_set_atime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "ctime", scm_entry );
    archive_entry_set_ctime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "mtime", scm_entry );
    archive_entry_set_mtime(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ), 0 );

    index = find_slot_index( "size", scm_entry );
    archive_entry_set_size(
        entry, Scm_GetInteger( Scm_InstanceSlotRef( scm_entry, index ) ) );
    index      = find_slot_index( "name", scm_entry );
    name_index = index;
    archive_entry_copy_pathname(
        entry, Scm_GetString(
                   SCM_STRING( Scm_InstanceSlotRef( scm_entry, index ) ) ) );
    /*
     * fprintf( stderr, "[%s]\n", Scm_GetString( SCM_STRING( Scm_InstanceSlotRef( scm_entry, index ) ) ) );
     * fprintf( stderr, "[%s]\n", archive_entry_pathname_utf8( entry ) );
     */

    // write header
    r = archive_write_header( aw, entry );
    if( r != ARCHIVE_OK ) {
      return AR_ERROR( r, archive_error_string( aw ), SCM_NIL );
    }

    if( isdir == 0 ) { /* file */
      index = find_slot_index( "type", scm_entry );
      sym   = Scm_InstanceSlotRef( scm_entry, index );
      index = find_slot_index( "data", scm_entry );
      oport = SCM_PORT( Scm_InstanceSlotRef( scm_entry, index ) );
      len   = SCM_DSTRING_SIZE( &( oport->src.ostr ) );

      if( COMPARE_SYMBOL( sym, "file" ) && len == 0 ) { /* header only */
        fd  = open( SCM_CSTR( Scm_InstanceSlotRef( scm_entry, name_index ) ),
                   O_RDONLY );
        len = read( fd, buf, BUFSIZ );
        while( len > 0 ) {
          if( archive_write_data( aw, buf, len ) != len ) {
            AR_ERROR( -1, "Size mismatch", SCM_NIL );
          }
          len = read( fd, buf, BUFSIZ );
        }
        close( fd );
      } else {
        if( archive_write_data( aw, Scm_DStringGetz( &( oport->src.ostr ) ),
                                len ) != len ) {
          return AR_ERROR( -1, "Size mismatch", SCM_NIL );
        }
      }
    }

    archive_entry_clear( entry );
  }
  archive_entry_free( entry );

  archive_write_close( aw );
  archive_write_free( aw );

  u8 = Scm_MakeUVector( SCM_CLASS_U8VECTOR, used, archive_buffer );

  return AR_ERROR( 0, "", u8 );
}
