#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(use gauche.collection)
(use util.archives)
(use gauche.uvector)

;;; Build archive from scratch.
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (let ((builder (open-build-archive-to-file "./b01.tar.xz"
                                            ARCHIVE-FORMAT-TAR-GNUTAR
                                            ARCHIVE-FILTER-XZ))
        (entry #f))
    (set! entry (create-entry-from-file "read_archive.scm"))
    (append-entry builder entry)

    ;;; create new archive entry on the memory.
    (set! entry (create-entry-from-memory "subdir/foo.txt"))
    ;;; Write data to entry.
    (format entry "text foo\n")
    (format entry "output string using format function\n")
    (append-entry builder entry)

    ;;; Create empty directory.
    (set! entry (create-entry-from-memory "emptydir/"))
    ;;; Don't forget to set entry slot 'isdir to `#t`.
    (set! (~ entry 'isdir) #t)
    (append-entry builder entry)

    (close-build-archive builder)))

;;; Following code is same as above it excepts building archive to memory.
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (let ((builder (open-build-archive-to-memory ARCHIVE-FORMAT-TAR-GNUTAR
                                               ARCHIVE-FILTER-XZ))
        (entry #f)
        (u8data #f))
    (set! entry (create-entry-from-file "read_archive.scm"))
    (append-entry builder entry)

    ;;; create new archive entry on the memory.
    (set! entry (create-entry-from-memory "subdir/foo.txt"))
    ;;; Write data to entry.
    (format entry "text foo\n")
    (format entry "output string using format function\n")
    (append-entry builder entry)

    ;;; Create empty directory.
    (set! entry (create-entry-from-memory "emptydir/"))
    ;;; Don't forget to set entry slot 'isdir to `#t`.
    (set! (~ entry 'isdir) #t)
    (append-entry builder entry)

    ;;; Get archive as u8vector.
    (set! u8data (close-build-archive builder))
    (call-with-output-file "./b02.tar.xz"
      (^(oport) (write-uvector u8data oport)))))
