#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(use gauche.collection)
(use util.archives)

(define-macro (read-entry archive)
  `(let ((entry (get-entry ,archive)))
    (format #t "~S\n" entry)
    (describe entry)))

;;; Read archive manually
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (let ((archive (open-input-archive-from-name "./testarchive01.tar")))
    (begin-iterator archive)
    (read-entry archive)
    (read-entry archive)
    (read-entry archive)
    (end-iterator archive)
    (close-input-archive archive)))

;;; Read archive with iterator protocol
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (let ((archive (open-input-archive-from-name "./testarchive01.tar")))
    (map (^(entry) (describe entry)) archive)
    (close-input-archive archive)))

;;; Same as above code
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (call-with-file-archive
   "./testarchive01.tar"
   (^(archive)
     (map (^(entry) (describe entry)) archive))))

;;; Read whole data from archive.
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (call-with-file-archive
   "./testarchive01.tar"
   (^(archive)
     (map
      (^(entry)
       ;;; Read data from archive via entry information.
       (read-data archive entry)
       (format #t "filename [~a], content [~a]\n"
               (~ entry 'name) (get-archive-entry-data entry)))
      archive))))


;;; Read archive from memory.
(guard
    (exc
     ((archive-error? exc)
      (format #t "~a ~a\n"
              (archive-error-debug-info exc)
              (archive-error-message exc))))
  (let ((mem-archive (open-output-string)))
    (call-with-input-file "./testarchive01.tar"
      (^(iport)
        (copy-port iport mem-archive)))
   (call-with-memory-archive
    mem-archive
   (^(archive)
     (map (^(entry) (describe entry)) archive)))))
